# coding: utf-8
# created: codeskyblue 2020/06/11
#
# ideviceinstaller -i .... ipa
# Copying 'WebDriverAgentRunner-Runner-resign.ipa' to device... DONE.
# Installing 'com.facebook.WebDriverAgentRunner.xctrunner'
#  - CreatingStagingDirectory (5%)
#  - ExtractingPackage (15%)
#  - InspectingPackage (20%)
#  - TakingInstallLock (20%)
#  - PreflightingApplication (30%)
#  - InstallingEmbeddedProfile (30%)
#  - VerifyingApplication (40%)
#  - CreatingContainer (50%)
#  - InstallingApplication (60%)
#  - PostflightingApplication (70%)
#  - SandboxingApplication (80%)
#  - GeneratingApplicationMap (90%)
#  - Complete

from pprint import pprint
from typing import Optional
import typing

from ._safe_socket import PlistSocket
from ._utils import logger


class Installation(PlistSocket):
    SERVICE_NAME = "com.apple.mobile.installation_proxy"

    def prepare(self):
        return super().prepare()

    def lookup(self, bundle_id: str) -> Optional[dict]:
        """
        Returns:
            LookupResult(dict) or None
        
        Protocol response:
        {'Status': 'Complete',
         'LookupResult': {
            'com.facebook.WebDriverAgentRunner.xctrunner': {
                'ApplicationType': 'User',
                'CFBundleIdentifier': 'com.facebook.WebDriverAgentRunner.xctrunner',
                'CFBundlePackageType': 'APPL',
                'CFBundleDisplayName': 'WebDriverAgentRunner-Runner',
                'CFBundleSignature': '????',
                'CFBundleInfoDictionaryVersion': '6.0',
                'CFBundleSupportedPlatforms': ['iPhoneOS'],
                'CFBundleNumericVersion': 16809984,
                'CFBundleName': 'WebDriverAgentRunner-Runner',
                'CFBundleShortVersionString': '1.0',
                'CFBundleExecutable': 'WebDriverAgentRunner-Runner',
                'CFBundleAllowMixedLocalizations': True,
                'CFBundleVersion': '1',
                'CFBundleDevelopmentRegion': 'en',
                'Container': '/private/var/mobile/Containers/Data/Application/7CA19F56-4CA2-40F5-A785-954174BD3AEF',
                'Path': '/private/var/containers/Bundle/Application/0A608C62-FBE7-43C0-B083-78F27AC5FF8E/WebDriverAgentRunner-Runner.app',
                'SignerIdentity': 'Apple Development: Shengxiang Sun (2YZG5Q7P9P)',
                'EnvironmentVariables': {'CFFIXED_USER_HOME': '/private/var/mobile/Containers/Data/Application/7CA19F56-4CA2-40F5-A785-954174BD3AEF',
                    'TMPDIR': '/private/var/mobile/Containers/Data/Application/7CA19F56-4CA2-40F5-A785-954174BD3AEF/tmp',
                    'HOME': '/private/var/mobile/Containers/Data/Application/7CA19F56-4CA2-40F5-A785-954174BD3AEF'},
                'BuildMachineOSBuild': '19E211',
                'MinimumOSVersion': '8.0',
                'LSRequiresIPhoneOS': True,
                'ProfileValidated': True,
                'SequenceNumber': 1624,
                'IsDemotedApp': False,
                'IsUpgradeable': True,
                'UIDeviceFamily': [1, 2],
                'UIRequiresFullScreen': True,
                'UIBackgroundModes': ['continuous'],
                'UIRequiredDeviceCapabilities': ['armv7'],
                'UISupportedInterfaceOrientations': ['UIInterfaceOrientationPortrait',
                    'UIInterfaceOrientationLandscapeLeft',
                    'UIInterfaceOrientationLandscapeRight'],
                'NSBluetoothAlwaysUsageDescription': 'Access is necessary for automated testing.',
                'NFCReaderUsageDescription': 'Access is necessary for automated testing.',
                'NSSiriUsageDescription': 'Access is necessary for automated testing.',
                'NSCameraUsageDescription': 'Access is necessary for automated testing.',
                'NSRemindersUsageDescription': 'Access is necessary for automated testing.',
                'NSHealthClinicalHealthRecordsShareUsageDescription': 'Access is necessary for automated testing.',
                'NSHealthUpdateUsageDescription': 'Access is necessary for automated testing.',
                'NSMotionUsageDescription': 'Access is necessary for automated testing.',
                'NSPhotoLibraryAddUsageDescription': 'Access is necessary for automated testing.',
                'NSHealthShareUsageDescription': 'Access is necessary for automated testing.',
                'NSAppleMusicUsageDescription': 'Access is necessary for automated testing.',
                'NSSpeechRecognitionUsageDescription': 'Access is necessary for automated testing.',
                'NSLocationUsageDescription': 'Access is necessary for automated testing.',
                'NSCalendarsUsageDescription': 'Access is necessary for automated testing.',
                'NSMicrophoneUsageDescription': 'Access is necessary for automated testing.',
                'NSLocationWhenInUseUsageDescription': 'Access is necessary for automated testing.',
                'NSFaceIDUsageDescription': 'Access is necessary for automated testing.',
                'NSLocationAlwaysAndWhenInUseUsageDescription': 'Access is necessary for automated testing.',
                'NSHomeKitUsageDescription': 'Access is necessary for automated testing.',
                'NSPhotoLibraryUsageDescription': 'Access is necessary for automated testing.',
                'NSAppTransportSecurity': {'NSAllowsArbitraryLoads': True},
                'Entitlements': {
                    'keychain-access-groups': ['M75PC2L4UP.com.facebook.WebDriverAgentRunner.xctrunner'],
                    'application-identifier': 'M75PC2L4UP.com.facebook.WebDriverAgentRunner.xctrunner',
                    'get-task-allow': True,
                    'com.apple.developer.team-identifier': 'M75PC2L4UP'},
                'DTPlatformVersion': '13.4',
                'DTSDKBuild': '17E218',
                'DTXcodeBuild': '11E605b',
                'DTPlatformName': 'iphoneos',
                'DTCompiler': 'com.apple.compilers.llvm.clang.1_0',
                'DTSDKName': 'iphoneos13.4.internal',
                'DTPlatformBuild': '17E218',
                'DTXcode': '1150'}
            }
        }
        """
        self.send_packet({
            "Command": "Lookup",
            "ClientOptions": {
                "BundleIDs": [bundle_id]
            }
        })
        ret = self.recv_packet()
        # Most used attributes
        # ApplicationType
        # CFBundleDisplayName
        # CFBundleExecutable
        # Path
        assert ret['Status'] == 'Complete'
        return ret['LookupResult'].get(bundle_id)

    def iter_installed(self, app_type: Optional[str] = "User", attrs: Optional[list]=None):
        """
        Args:
            app_type (str): one of ['User', 'System']
            attrs: list
        
        Example attrs:
            ['ApplicationType',
            'CFBundleDisplayName',
            'CFBundleExecutable',
            "CFBundleIdentifier",
            'CFBundleName',
            'CFBundleShortVersionString',
            'CFBundleVersion',
            'Container',
            'Entitlements',
            'EnvironmentVariables',
            'MinimumOSVersion',
            'Path',
            'ProfileValidated',
            'SBAppTags',
            'SignerIdentity',
            'UIDeviceFamily',
            'UIRequiredDeviceCapabilities']
        
        Example protocol response:
            {'Status': 'BrowsingApplications',
            'CurrentAmount': 9,
            'CurrentIndex': 0,
            'CurrentList': [{'ApplicationType': 'User',
                            'CFBundleDisplayName': 'Demo应用',
                            'CFBundleIdentifier': 'com.example.demo',
                            等等
                            },...]
        }
        """
        options = {}
        if app_type:
            options["ApplicationType"] = app_type
        if attrs:
            options['ReturnAttributes'] = attrs
        # options['ShowLaunchProhibitedApps'] = True 

        self.send_packet({
            "Command": "Browse",
            "ClientOptions": options,
        })
        # })
        while True:
            data = self.recv_packet()
            if data['Status'] == 'Complete':
                break
            for appinfo in data['CurrentList']:
                yield appinfo





#!/usr/bin/python
# This is a simple port-forward / proxy, written using only the default python
# library. If you want to make a suggestion or fix something you can contact-me
# at voorloop_at_gmail.com
# Distributed over IDC(I Don't Care) license

# Python3
# Modified: 2020/05/15 (Fri) shengxiang.ssx

import argparse
import datetime
import os
import plistlib
import pprint
import re
import select
import socket
import ssl
import string
import struct
import sys
import threading
import time
import traceback

import hexdump
from logzero import logger

# Changing the buffer_size and delay, you can improve the speed and bandwidth.
# But when buffer get to high or delay go too down, you can broke things
buffer_size = 40960
delay = 0.0001


_package_index = [0]

def next_package_index() -> int:
    _package_index[0] += 1
    return _package_index[0]


def create_socket(addr) -> socket.socket:
    if isinstance(addr, (tuple, list)):
        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    elif isinstance(addr, str):
        return socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)


def is_ssl_data(data: bytes) -> bool:
    """ FIXME(ssx): better to change to EnableSSLSession """
    return len(data) >= 8 and \
        data[:3] == b'\x16\x03\x01' and data[5:6] == b'\x01'


def recvall(sock: socket.socket, n: int) -> bytearray:
    buf = bytearray()
    while n > len(buf):
        chunk = sock.recv(n - len(buf))
        if not chunk:
            raise ValueError("socket not recv all bytes")
        buf.extend(chunk)
    return buf


class TheServer:
    input_list = []
    channel = {}
    socket_tags = {}

    def __init__(self,
                 listen_addr: str,
                 forward_to: str,
                 parse_ssl: bool = False,
                 pemfile=None):
        print("Listening on", listen_addr, "forward to", forward_to)
        self.server = create_socket(listen_addr)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind(listen_addr)
        if isinstance(listen_addr, str):
            os.chmod(listen_addr, 0o777)
        self.server.listen(200)

        self.ssl_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.ssl_server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.ssl_server.bind(('localhost', 10443))
        self.ssl_server.listen(200)

        self.forward_to = forward_to
        self.parse_ssl = parse_ssl
        self.pemfile = pemfile

        self.__tempsocks = {}  # Store SSLSocket
        self.__tag = 0
        self._data_directory = "usbmuxd-dumpdata/" + time.strftime(
            "%Y%m%d-%H-%M-%S")

    def main_loop(self):
        self.input_list.append(self.server)
        self.input_list.append(self.ssl_server)

        while True:
            time.sleep(delay)
            timeout = .5
            inputready, outputready, exceptready = select.select(self.input_list, [], [], timeout)
            for self.s in inputready:
                if self.s == self.server:
                    self.on_accept()
                    break

                if self.s == self.ssl_server:
                    self.on_ssl_accept()
                    break

                try:
                    self.data = self.s.recv(buffer_size)
                    if len(self.data) == 0:
                        self.on_close()
                        break
                    self.on_recv()
                except ssl.SSLError as e:
                    logger.warning("SSLError: tag: %d, %s",
                                   self.socket_tags[self.s], e)

                    self.on_close()
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    #logger.warning("Unknown error: %s", e)

    def gen_tag(self) -> int:  # return uniq int
        self.__tag += 1
        return self.__tag

    def pipe_socket(self, clientsock, serversock, tag=None):
        assert clientsock not in self.channel, (clientsock, "already piped")
        self.input_list.append(clientsock)
        self.input_list.append(serversock)
        self.channel[clientsock] = serversock
        self.channel[serversock] = clientsock
        if tag:
            self.socket_tags[serversock] = -tag
            self.socket_tags[clientsock] = tag

    def unpipe_socket(self, clientsock) -> int:
        """
        return socket tag
        """
        serversock = self.channel[clientsock]
        del self.channel[clientsock]
        del self.channel[serversock]
        self.remove_from_list(self.input_list, clientsock)
        self.remove_from_list(self.input_list, serversock)
        self.socket_tags.pop(serversock, None)  # socket_tags
        return self.socket_tags.pop(clientsock, None)

    def remove_from_list(self, _list, value):
        try:
            _list.remove(value)
        except ValueError:
            pass

    def on_accept(self):
        forward = create_socket(self.forward_to)
        forward.connect(self.forward_to)

        clientsock, clientaddr = self.server.accept()
        if forward:
            print('[proxy]', clientaddr, "has connected")
            self.pipe_socket(clientsock, forward, self.gen_tag())
        else:
            print("[proxy] Can't establish connection with remote server.", )
            print("[proxy] Closing connection with client side", clientaddr)
            clientsock.close()

    def on_ssl_accept(self):
        sock, addr = self.ssl_server.accept()

        header = recvall(sock, 4)  #sock.recv(4)
        (tag, ) = struct.unpack("I", header)
        logger.info("on ssl accept: %s, tag: %d", addr, tag)
        ssl_serversock = self.__tempsocks[tag]

        def wait_ssl_socket():
            ssock = ssl.wrap_socket(sock,
                                    keyfile=self.pemfile,
                                    certfile=self.pemfile,
                                    server_side=True)
            self.pipe_socket(ssock, ssl_serversock, tag)

        th = threading.Thread(target=wait_ssl_socket)
        th.daemon = True
        th.start()

    def pretty_format_data(self, data: bytes) -> str:
        try:
            return '\n'.join(list(self._iter_pretty_format_data(data)))
        except:
            print(list(self._iter_pretty_format_data(data)))
            raise

    def _iter_pretty_format_data(self, data: bytes):
        if b'<plist version=' in data:
            lindex = data.find(b'<?xml version=')
            rindex = data.find(b'</plist>') + len('</plist>')
            plistdata = data[lindex:rindex]

            yield hexdump.hexdump(data[:lindex], "return")
            yield "## Plist-XML"
            try:
                pdata = plistlib.loads(plistdata)
                if 'PairRecordData' in pdata:
                    yield "... PairRecordData ..."
                else:
                    yield pprint.pformat(pdata)
                yield hexdump.hexdump(data[rindex:], "return")
            except:
                yield hexdump.hexdump(data, "return")
        elif b'bplist00' in data:
            # parse bplist data
            from tinstruments import bplist
            buf = data
            while True:
                lindex = buf.find(b'bplist00')
                #print("bplist00", hex(lindex))
                if lindex == -1:
                    break
                m = re.search(b"\x00{6}[\x02\x01]{2}.{24}", buf)
                if not m:
                    break
                rindex = m.end()

                yield hexdump.hexdump(buf[:lindex], "return")
                try:
                    yield "## NSKeyedArchiver-BINARY: [{}, {}]".format(hex(lindex), hex(rindex))
                    pdata = bplist.objc_decode(buf[lindex:rindex])
                    yield pprint.pformat(pdata)
                except bplist.InvalidNSKeyedArchiverFormat:
                    yield "## Plist-BINARY: [{}, {}]".format(hex(lindex), hex(rindex))
                    pdata = bplist.loads(buf[lindex:rindex])
                    yield pprint.pformat(pdata)
                except Exception as e:
                    yield "## Parse plist-BINARY error: [{}, {}]".format(hex(lindex), hex(rindex))
                    yield traceback.format_exc()
                    # BPlistdata failed string indices must be integers
                    yield "load binary plistdata failed: {}".format(e)
                    yield hexdump.hexdump(buf[lindex:rindex], "return")
                finally:
                    buf = buf[rindex:]
            yield hexdump.hexdump(buf, "return")
        else:  # just call hexdump
            yield f"## RAW length={len(data)}"
            max_length = 512
            if len(data) > max_length // 2:
                yield hexdump.hexdump(data[:max_length // 2], "return")
                yield "........ " * 5
                yield hexdump.hexdump(data[-max_length // 2:], "return")
            else:
                yield hexdump.hexdump(data, "return")
        # else:
        #     try:
        #         pdata = data.decode('utf-8')
        #         print(pdata)
        #     except UnicodeDecodeError:
        #         hexdump.hexdump(data)

    def dump_data(self, data):
        if self.s not in self.socket_tags:
            return

        tag = self.socket_tags[self.s]
        direction = ">"
        if tag < 0:
            direction = "<"
        index = abs(tag)

        print(f' {index} '.center(50, direction))
        if isinstance(self.s, ssl.SSLSocket):  # secure tunnel
            print('\33[32m', end="")
        print(f"Length={len(data)} 0x{len(data):02X}")
        print("Time:", datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3])

        # TODO
        #if tag < 0:
        #    print("Skip show receving data")
        #    return
        if True:
            os.makedirs(self._data_directory, 0o755, exist_ok=True)
            fpath = os.path.join(self._data_directory, f"{index}.txt")
            with open(fpath, "a") as f:
                f.write(f'# INDEX: {index} {direction*5}\n')
                f.write(f"Size: oct:{len(data)} hex:0x{len(data):02X}\n")
                f.write(f"Time: {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]}\n")
                f.write(f"Package index: {next_package_index()}\n")
                f.write(self.pretty_format_data(data) + "\n")
                f.write('\n\n')

        if True:
            for line in self._iter_pretty_format_data(data):
                print(line)

        #else:
        #    try:
        #        pdata = data.decode('utf-8')
        #        print(pdata)
        #    except UnicodeDecodeError:
        #        hexdump.hexdump(data)

        if isinstance(self.s, ssl.SSLSocket):  # secure tunnel
            print('\33[0m', end="")

    def man_in_middle_ssl(self, clientsock, ssl_hello_data: bytes):
        serversock = self.channel[clientsock]
        tag = self.unpipe_socket(clientsock)

        ssl_serversock = ssl.wrap_socket(
            serversock,
            keyfile=self.pemfile,  # iPhoneSE
            certfile=self.pemfile)
        print("serversock secure ready, tag: {}".format(tag))

        self.__tempsocks[tag] = ssl_serversock

        mim_clientsock = socket.create_connection(('localhost', 10443))
        mim_clientsock.sendall(struct.pack("I", tag))

        hexdump.hexdump(ssl_hello_data[:1024])
        mim_clientsock.sendall(ssl_hello_data)
        self.pipe_socket(clientsock, mim_clientsock)

    def on_recv(self):
        # here we can parse and/or modify the data before send forward
        data = self.data

        # Check SSL ClientHello message
        if self.parse_ssl and is_ssl_data(data):
            logger.info("Mock SSL-Server")
            clientsock = self.s
            self.man_in_middle_ssl(clientsock, data)
            return

        # if b'PairRecordData' in data and b'DeviceCertificate' in data:
        #     print(".... PairRecordData ....")
        # else:
        self.dump_data(data)

        try:
            self.channel[self.s].send(data)
        except BrokenPipeError:
            print("channel broken pipe")

    def on_close(self):
        print('[proxy]', self.socket_tags.get(self.s), "has disconnected")
        # print('[proxy]', self.channel[self.s].getpeername(), "has disconnected, too")

        #remove objects from input_list

        # self.input_list.remove(self.s)
        # self.input_list.remove(self.channel[self.s])

        out = self.channel[self.s]
        # close the connection with client
        self.channel[out].close()  # equivalent to do self.s.close()
        # close the connection with remote server
        self.channel[self.s].close()
        # delete both objects from channel dict

        self.unpipe_socket(self.s)
        # del self.channel[out]
        # del self.channel[self.s]


def _parse_addr(addr: str):
    if ':' in addr:
        host, port = addr.split(":", 1)
        return (host, int(port))
    return addr


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-L",
                        "--listen-addr",
                        default="/var/run/usbmuxd",
                        help="listen address, eg :5555 or /tmp/listen.sock")
    parser.add_argument(
        "-F",
        "--forward-to",
        default="/var/run/usbmuxx",
        help="forward to, eg: localhost:5037 or /var/lib/usbmuxd")
    parser.add_argument("-S",
                        "--ssl",
                        action="store_true",
                        help="parse ssl data")
    parser.add_argument("--pemfile", help="ssl pemfile")
    args = parser.parse_args()
    #print(args)

    listen_addr = _parse_addr(args.listen_addr)
    forward_to = _parse_addr(args.forward_to)

    server = TheServer(listen_addr,
                       forward_to,
                       parse_ssl=args.ssl,
                       pemfile=args.pemfile)
    try:
        server.main_loop()
    except KeyboardInterrupt:
        print("Ctrl C - Stopping server")
        sys.exit(1)
    finally:
        if isinstance(listen_addr, str):
            import os
            os.unlink(listen_addr)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Created on Mon Jan 04 2021 17:22:26 by codeskyblue
"""

import argparse
import base64
import json
import logging
import os
import re
import subprocess
import sys
import time
from collections import defaultdict
from pprint import pformat, pprint
from typing import Optional, Union
from datetime import datetime

import requests
from logzero import setup_logger

from ._device import Device
from ._imagemounter import cache_developer_image
from ._ipautil import IPAReader
from ._perf import DataType
from ._proto import LOG, MODELS, PROGRAM_NAME
from ._relay import relay
from ._usbmux import Usbmux
from ._utils import get_app_dir, get_binary_by_name, is_atty
from ._version import __version__
from ._wdaproxy import WDAService
from .exceptions import MuxError, MuxServiceError, ServiceError

um = None  # Usbmux
logger = logging.getLogger(PROGRAM_NAME)


def _complete_udid(udid: Optional[str] = None) -> str:
    infos = um.device_list()
    if not udid:
        if len(infos) >= 2:
            sys.exit("More than 2 devices detected")
        if len(infos) == 0:
            sys.exit("No device detected")
        return infos[0].udid

    # Find udid exactly match
    for info in infos:
        if info.udid == udid:
            return udid

    # Find udid starts-with
    _udids = [
        info.udid for info in infos
        if info.udid.startswith(udid)
    ]

    if len(_udids) == 1:
        return _udids[0]

    raise RuntimeError("No matched device", udid)


def _udid2device(udid: Optional[str] = None) -> Device:
    _udid = _complete_udid(udid)
    if _udid != udid:
        logger.debug("AutoComplete udid %s", _udid)
    del (udid)
    return Device(_udid, um)


def cmd_list(args: argparse.Namespace):
    if not args.json and is_atty:
        print("List of apple devices attached", file=sys.stderr)

    result = []
    for dinfo in um.device_list():
        udid, conn_type = dinfo.udid, dinfo.conn_type
        try:
            _d = Device(udid, um)
            name = _d.name
            if not args.json:
                print(udid, name, conn_type)
        except MuxError:
            name = ""
        result.append(dict(udid=udid, name=name, conn_type=conn_type))
    if args.json:
        _print_json(result)


def _print_json(value):
    def _bytes_hook(obj):
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode()
        else:
            return str(obj)

    print(json.dumps(value, indent=4, ensure_ascii=False, default=_bytes_hook))


def cmd_device_info(args: argparse.Namespace):
    d = _udid2device(args.udid)
    value = d.get_value(no_session=args.simple,
                        key=args.key,
                        domain=args.domain)
    if args.json:
        _print_json(value)
    elif args.key or args.domain:
        pprint(value)
    else:
        print("{:17s} {}".format("MarketName:",
                                 MODELS.get(value['ProductType'])))
        for attr in ('DeviceName', 'ProductVersion', 'ProductType',
                     'ModelNumber', 'SerialNumber', 'PhoneNumber',
                     'CPUArchitecture', 'ProductName', 'ProtocolVersion',
                     'RegionInfo', 'TimeIntervalSince1970', 'TimeZone',
                     'UniqueDeviceID', 'WiFiAddress', 'BluetoothAddress',
                     'BasebandVersion'):
            print("{:17s} {}".format(attr + ":", value.get(attr)))


def cmd_date(args: argparse.Namespace):
    d = _udid2device(args.udid)
    value = d.get_value() or {}
    timestamp = value.get("TimeIntervalSince1970")
    if args.format:
        print(datetime.fromtimestamp(int(timestamp)))
    else:
        print(timestamp)


def cmd_version(args: argparse.Namespace):
    print(PROGRAM_NAME, "version", __version__)


def cmd_install(args: argparse.Namespace):
    d = _udid2device(args.udid)
    bundle_id = d.app_install(args.filepath_or_url)

    if args.launch:
        pid = d.instruments.app_launch(bundle_id)
        logger.info("Launch %r, process pid: %d", bundle_id, pid)


def cmd_uninstall(args: argparse.Namespace):
    d = _udid2device(args.udid)
    ok = d.app_uninstall(args.bundle_id)
    if not ok:
        sys.exit(1)


def cmd_reboot(args: argparse.Namespace):
    d = _udid2device(args.udid)
    print(d.reboot())


def cmd_shutdown(args: argparse.Namespace):
    d = _udid2device(args.udid)
    print(d.shutdown())


def cmd_parse(args: argparse.Namespace):
    uri = args.uri
    _all = args.all

    fp = None
    if re.match(r"^https?://", uri):
        try:
            import httpio
        except ImportError:
            print("Install missing lib: httpip")
            retcode = subprocess.call(
                [sys.executable, '-m', 'pip', 'install', '-U', 'httpio'])
            assert retcode == 0
            import httpio

        fp = httpio.open(uri, block_size=-1)
    else:
        assert os.path.isfile(uri)
        fp = open(uri, 'rb')

    try:
        ir = IPAReader(fp)
        ir.dump_info(all=_all)
    finally:
        fp.close()


def cmd_watch(args: argparse.Namespace):
    """
    Info example:
    {'DeviceID': 13,
     'MessageType': 'Attached',
     'Properties': {'ConnectionSpeed': 480000000,
                    'ConnectionType': 'USB',
                    'DeviceID': 13,
                    'LocationID': 340918272,
                    'ProductID': 4776,
                    'SerialNumber': '84ad172e22d8372eb752f413280722cdcc200954',
                    'USBSerialNumber': '84ad172e22d8372eb752f413280722cdcc200954'}}
    """
    for info in um.watch_device():
        logger.info("%s", pformat(info))


def cmd_wait_for_device(args):
    u = Usbmux(args.socket)

    for info in u.watch_device():
        logger.debug("%s", pformat(info))
        if info['MessageType'] != 'Attached':
            continue
        udid = info['Properties']['SerialNumber']
        if args.udid is None:
            break
        if udid == args.udid:
            print("Device {!r} attached".format(
                info['Properties']['SerialNumber']))
            break


def cmd_xctest(args: argparse.Namespace):
    """
    Run XCTest required WDA installed.
    """
    if args.debug:
        setup_logger(LOG.xctest, level=logging.DEBUG)

    d = _udid2device(args.udid)
    env = {}
    for kv in args.env or []:
        key, val = kv.split(":", 1)
        env[key] = val
    if env:
        logger.info("Launch env: %s", env)
    d.xctest(args.bundle_id,
             target_bundle_id=args.target_bundle_id,
             logger=setup_logger(level=logging.INFO),
             env=env)


def cmd_screenshot(args: argparse.Namespace):
    d = _udid2device(args.udid)
    filename = args.filename or "screenshot.jpg"
    print("Screenshot saved to", filename)
    d.screenshot().convert("RGB").save(filename)


def cmd_app_info(args: argparse.Namespace):
    d = _udid2device(args.udid)
    info = d.installation.lookup(args.bundle_id)
    if info is None:
        sys.exit(1)
    if args.json:
        _print_json(info)
    else:
        pprint(info)


def cmd_applist(args: argparse.Namespace):
    d = _udid2device(args.udid)
    # appinfos = d.installation.list_installed()
    # apps = d.instruments.app_list()
    # pprint(apps)
    for info in d.installation.iter_installed():
        # bundle_path = info['BundlePath']
        bundle_id = info['CFBundleIdentifier']

        try:
            display_name = info['CFBundleDisplayName']
            # major.minor.patch
            version = info.get('CFBundleShortVersionString', '')
            print(bundle_id, display_name, version)
        except BrokenPipeError:
            break
        # print(" ".join(
        #     (bundle_path, bundle_id, info['DisplayName'],
        #         info.get('Version', ''), info['Type'])))


def cmd_launch(args: argparse.Namespace):
    d = _udid2device(args.udid)
    try:
        pid = d.instruments.app_launch(args.bundle_id,
                                       args=args.arguments,
                                       kill_running=args.kill)
        print("PID:", pid)
    except ServiceError as e:
        sys.exit(e)


def cmd_kill(args: argparse.Namespace):
    d = _udid2device(args.udid)
    if args.name.isdigit():
        pid = int(args.name)
        d.app_kill(int(args.name))
    else:
        pid = d.app_kill(args.name)
        if pid is None:
            print("No app killed")
        else:
            print("Kill pid:", pid)


def cmd_system_info(args):
    d = _udid2device(args.udid)
    sinfo = d.instruments.system_info()
    pprint(sinfo)


def cmd_battery(args: argparse.Namespace):
    d = _udid2device(args.udid)
    power_info = d.get_io_power()
    if args.json:
        _print_json(power_info)
    else:
        pprint(power_info)


def cmd_developer(args: argparse.Namespace):
    if args.download_all:
        for major in range(7, 15):
            for minor in range(0, 10):
                version = f"{major}.{minor}"
                try:
                    cache_developer_image(version)
                except requests.HTTPError:
                    break
        #     print("finish cache developer image {}".format(version))
    else:
        d = _udid2device(args.udid)
        d.mount_developer_image()
    return


def cmd_relay(args: argparse.Namespace):
    d = _udid2device(args.udid)
    relay(d, args.lport, args.rport, debug=args.x)


def cmd_wdaproxy(args: argparse.Namespace):
    """ start xctest and relay """
    d = _udid2device(args.udid)
    
    env = {}
    for kv in args.env or []:
        key, val = kv.split(":", 1)
        env[key] = val
    if env:
        logger.info("Launch env: %s", env)
    
    serv = WDAService(d, args.bundle_id, env)
    serv.set_check_interval(args.check_interval)

    p = None
    if args.port:
        cmds = [
            sys.executable, '-m', 'tidevice', '-u', d.udid, 'relay',
            str(args.port), '8100'
        ]
        p = subprocess.Popen(cmds, stdout=sys.stdout, stderr=sys.stderr)

    try:
        serv.start()
        while serv._service.running and p.poll() is None:
            time.sleep(.1)
    finally:
        p and p.terminate()
        serv.stop()


def cmd_syslog(args: argparse.Namespace):
    d = _udid2device(args.udid)
    s = d.start_service("com.apple.syslog_relay")
    # print("SS")
    try:
        while True:
            text = s.recv().decode('utf-8')
            print(text, end='', flush=True)
    except (BrokenPipeError, IOError):
        # Python flushes standard streams on exit; redirect remaining output
        # to devnull to avoid another BrokenPipeError at shutdown
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())


def cmd_dump_fps(args):
    d = _udid2device(args.udid)
    for data in d.instruments.iter_opengl_data():
        if isinstance(data, str):
            continue
        fps = data['CoreAnimationFramesPerSecond']
        print("{:>2d} {}".format(fps, "-" * fps))


def cmd_pair(args: argparse.Namespace):
    d = _udid2device(args.udid)
    pair_record = d.pair()
    print("Paired with device", d.udid, "HostID:", pair_record['HostID'])


def cmd_fsync(args: argparse.Namespace):
    d = _udid2device(args.udid)
    if args.bundle_id:
        sync = d.app_sync(args.bundle_id)
    else:
        sync = d.sync

    arg0 = args.arguments[0]
    if args.command == 'ls':
        pprint(sync.listdir(arg0))
    elif args.command == 'rm':
        for arg in args.arguments:
            pprint(sync.remove(arg))
    elif args.command == 'stat':
        finfo = sync.stat(arg0)
        print("Fmt:", finfo.st_ifmt)
        print("CTime:", finfo.st_ctime)
        print("MTime:", finfo.st_mtime)
        print("Size:", finfo.st_size)
    elif args.command == 'tree':
        sync.treeview(arg0, depth=-1)
    elif args.command == 'pull':
        data = sync.pull_content(arg0)
        with open(os.path.basename(arg0), 'wb') as f:
            f.write(data)
        print("pulled to", os.path.basename(arg0))
    elif args.command == 'cat':
        for chunk in sync.iter_content(arg0):
            sys.stdout.write(chunk.decode('utf-8'))
            sys.stdout.flush()
    elif args.command == 'push':
        local_path = args.arguments[0]
        device_path = args.arguments[1]
        assert os.path.isfile(local_path)
        with open(local_path, "rb") as f:
            content = f.read()
            sync.push_content(device_path, content)
            print("pushed to", device_path)
    elif args.command == 'rmtree':
        pprint(sync.rmtree(arg0))
    elif args.command == 'mkdir':
        pprint(sync.mkdir(arg0))
    else:
        raise NotImplementedError()


def cmd_ps(args: argparse.Namespace):
    d = _udid2device(args.udid)
    app_infos = list(d.installation.iter_installed(app_type=None))
    ps = list(d.instruments.app_process_list(app_infos))

    lens = defaultdict(int)
    json_data = []
    keys = ['pid', 'name', 'bundle_id', 'display_name']
    for p in ps:
        if not args.all and not p['isApplication']:
            continue
        for key in keys:
            lens[key] = max(lens[key], len(str(p[key])))
        json_data.append({key: p[key] for key in keys})

    if args.json:
        _print_json(json_data)
        return

    # {:0} is not allowed, so max(1, xx) is necessary
    fmt = ' '.join(['{:%d}' % max(1, lens[key]) for key in keys])
    fmt = '{:>' + fmt[2:]  # set PID right align
    if is_atty:
        print(fmt.format(*[key.upper() for key in keys]))

    for p in ps:
        if not args.all and not p['isApplication']:
            continue
        print(fmt.format(*[p[key] for key in keys]))


def cmd_perf(args: argparse.Namespace):
    assert args.bundle_id
    #print("BundleID:", args.bundle_id)
    from ._perf import Performance
    d = _udid2device(args.udid)
    perfs = list(DataType)
    if args.perfs:
        perfs = []
        for _typename in args.perfs.split(","):
            perfs.append(DataType(_typename))
    # print(perfs)
    perf = Performance(d, perfs=perfs)

    def _cb(_type: DataType, data):
        print(_type.value, data)

    try:
        perf.start(args.bundle_id, callback=_cb)
        #print("Ctrl-C to finish")
        while True:
            time.sleep(.1)
    finally:
        perf.stop()


def cmd_test(args: argparse.Namespace):
    print("Just test")
    # files = os.listdir(path)

    # Here need device unlocked
    # signatures = d.imagemounter.lookup()
    # if signatures:
    #     logger.info("DeveloperImage already mounted")
    #     return


_commands = [
    dict(action=cmd_version, command="version", help="show current version"),
    dict(action=cmd_list,
         command="list",
         flags=[
             dict(args=['--json'],
                  action='store_true',
                  help='output in json format')
         ],
         help="show connected iOS devices"),
    dict(
        action=cmd_device_info,
        command="info",
        flags=[
            dict(args=['--json'],
                 action='store_true',
                 help="output as json format"),
            dict(
                args=['-s', '--simple'],
                action='store_true',
                help=
                'use a simple connection to avoid auto-pairing with the device'
            ),
            dict(args=['-k', '--key'],
                 type=str,
                 help='only query specified KEY'),
            dict(args=['--domain'], help='set domain of query to NAME.'),
        ],
        help="show device info"),
    dict(action=cmd_date,
         command="date",
         flags=[
             dict(args=['--format'],
                  action='store_true',
                  help="format timestamp")
         ],
         help="device current date"),
    dict(action=cmd_system_info,
         command="sysinfo",
         help="show device system info"),
    dict(action=cmd_app_info,
         command="appinfo",
         flags=[
             dict(args=['--json'],
                  action='store_true',
                  help='format output as json'),
             dict(args=['bundle_id'], help='bundle identifier'),
         ],
         help="inspect app info"),
    dict(action=cmd_applist, command="applist", help="list packages"),
    dict(action=cmd_battery,
         command='battery',
         flags=[
             dict(args=['--json'],
                  action='store_true',
                  help='format output as json')
         ],
         help='show battery info'),
    dict(action=cmd_screenshot,
         command="screenshot",
         help="take screenshot",
         flags=[dict(args=['filename'], nargs="?", help="output filename")]),
    dict(action=cmd_install,
         command="install",
         flags=[
             dict(args=['-L', '--launch'],
                  action='store_true',
                  help='launch after installed'),
             dict(args=['filepath_or_url'], help="local filepath or url")
         ],
         help="install application"),
    dict(action=cmd_uninstall,
         command="uninstall",
         flags=[dict(args=['bundle_id'], help="bundle_id of application")],
         help="uninstall application"),
    dict(action=cmd_reboot, command="reboot", help="reboot device"),
    dict(action=cmd_shutdown, command="shutdown", help="shutdown device"),
    dict(action=cmd_parse,
         command="parse",
         flags=[dict(args=['--all'], action='store_true', help='show all info'),
                dict(args=['uri'], help="local path or url")],
         help="parse ipa bundle id"),
    dict(action=cmd_watch, command="watch", help="watch device"),
    dict(action=cmd_wait_for_device,
         command='wait-for-device',
         help='wait for device attached'),
    dict(action=cmd_launch,
         command="launch",
         flags=[
             dict(args=['--kill'],
                  action='store_true',
                  help='kill app if running'),
             dict(args=["bundle_id"], help="app bundleId"),
             dict(args=['arguments'], nargs='*', help='app arguments'),
         ],
         help="launch app with bundle_id"),
    dict(action=cmd_kill,
         command="kill",
         flags=[dict(args=['name'], help='pid or bundle_id')],
         help="kill by pid or bundle_id"),
    dict(action=cmd_ps,
         command="ps",
         flags=[
             dict(args=['--json'],
                  action='store_true',
                  help='format output as json'),
             dict(args=['-A', '--all'],
                  action='store_true',
                  help='show all process')
         ],
         help="show running processes"),
    dict(action=cmd_relay,
         command="relay",
         flags=[
             dict(args=['-x'],
                  action='store_true',
                  help='verbose data traffic, hexadecimal'),
             dict(args=['lport'], type=int, help='local port'),
             dict(args=['rport'], type=int, help='remote port'),
         ],
         help="relay phone inner port to pc, same as iproxy"),
    dict(
        action=cmd_xctest,
        command="xctest",
        flags=[
            dict(args=['--debug'], action='store_true', help='show debug log'),
            dict(args=['-B', '--bundle_id', '--bundle-id'],
                 default="com.facebook.*.xctrunner",
                 help="bundle id of the test to launch"),
            dict(args=['--target-bundle-id'],
                 help='bundle id of the target app [optional]'),
            #  dict(args=['-I', '--install-wda'],
            #       action='store_true',
            #       help='install webdriveragent app'),
            dict(args=['-e', '--env'],
                 action='append',
                 help="set env with format key:value, support multi -e"),
        ],
        help="run XCTest"),
    dict(action=cmd_wdaproxy,
         command='wdaproxy',
         flags=[
             dict(args=['-B', '--bundle_id'],
                  default="com.facebook.*.xctrunner",
                  help="test application bundle id"),
             dict(args=['-p', '--port'],
                  type=int,
                  default=8100,
                  help='pc listen port, set to 0 to disable port forward'),
             dict(args=['-e', '--env'],
                  action='append',
                  help="set env with format key:value, support multi -e"),
             dict(args=['--check-interval'], type=float, default=30.0,
                  help="check if wda is alive every CHECK_INTERVAL seconds, stop check when set to 0"),
         ],
         help='keep WDA running and relay WDA service to pc'),
    dict(action=cmd_syslog, command='syslog', help="print iphone syslog"),
    dict(action=cmd_fsync,
         command="fsync",
         flags=[
             dict(args=['-B', '--bundle_id'], help='app bundle id'),
             dict(args=['command'],
                  choices=[
                      'ls', 'rm', 'cat', 'pull', 'push', 'stat', 'tree',
                      'rmtree', 'mkdir'
                  ]),
             dict(args=['arguments'], nargs='+', help='command arguments'),
         ],
         help="app file management"),
    dict(action=cmd_dump_fps, command='dumpfps', help='dump fps'),
    dict(action=cmd_developer,
         command="developer",
         flags=[dict(args=['--download-all'], action="store_true", help="download all developer to local")],
         help="mount developer image to device"),
    dict(action=cmd_pair, command='pair', help='pair device'),
    dict(action=cmd_perf,
         command="perf",
         flags=[dict(args=['-B', '--bundle_id'], help='app bundle id', required=True),
                dict(args=['-o'], dest='perfs', help='cpu,memory,fps,network,screenshot. separate by ","', required=False),],
         help="performance of app"),
    dict(action=cmd_test, command="test", help="command for developer"),
]


def main():
    # yapf: disable
    parser = argparse.ArgumentParser(
        description="Tool for communicate with iOS devices, version {}, created: codeskyblue 2020/05".format(__version__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-v", "--version", action="store_true", help="show current version"),
    parser.add_argument("-u", "--udid", help="specify unique device identifier")
    parser.add_argument("--socket", help="usbmuxd listen address, host:port or local-path")

    subparser = parser.add_subparsers(dest='subparser')
    actions = {}
    for c in _commands:
        cmd_name = c['command']
        actions[cmd_name] = c['action']
        sp = subparser.add_parser(cmd_name, help=c.get('help'),
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        for f in c.get('flags', []):
            args = f.get('args')
            if not args:
                args = ['-'*min(2, len(n)) + n for n in f['name']]
            kwargs = f.copy()
            kwargs.pop('name', None)
            kwargs.pop('args', None)
            sp.add_argument(*args, **kwargs)

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return

    if not args.subparser:
        parser.print_help()
        # show_upgrade_message()
        return

    # log setup
    setup_logger(LOG.main,
        level=logging.DEBUG if os.getenv("DEBUG") in ("1", "on", "true") else logging.INFO)

    global um
    um = Usbmux(args.socket)
    actions[args.subparser](args)
    # yapf: enable


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Created on Fri Sep 24 2021 14:13:00 by codeskyblue
"""

import typing
import enum


class ConnectionType(str, enum.Enum):
    USB = "usb"
    NETWORK = "network"


class _BaseInfo:
    def _asdict(self) -> dict:
        """ for simplejson """
        return self.__dict__.copy()
    
    def __repr__(self) -> str:
        attrs = []
        for k, v in self.__dict__.items():
            attrs.append(f"{k}={v!r}")
        return f"<{self.__class__.__name__} " + ", ".join(attrs) + ">"


class DeviceInfo(_BaseInfo):
    udid: str
    device_id: int
    conn_type: ConnectionType
